package	LinkBanner;

=head1 NAME

LinkBanner.pm - Respond for CGI as the link banner image or the links page of referers

$Id: LinkBanner.pm,v 1.6 2002/04/11 15:06:57 T.Shigetome Exp $

=head1 SYNOPSIS

This will respond a banner image and will log referer address.
That referer information is used for to make links list.

	use LinkBanner;
	my($banner) = new LinkBanner("/home/foo/public_html/log_path/");
	$banner->RecordLog();
	$banner->OutBanner("/home/foo/public_html/cgi-bin/testbanner.jpg");


	use LinkBanner;
	my($banner) = new LinkBanner("/home/foo/public_html/log_path/");
	$banner->deny("/home/foo/public_html/deny_list.txt");
	$banner->OutLinkPage();

=head1 DESCRIPTION

This will automatically create link page.
If some web contents reference to CGI program is useing this module, that will respond banner image.
At that time also this module will be logging address of reference page.
When required links page, this will generate from that logged file.

=head2 METHOD

=over 4

=item new([C<$logpath> ,C<$query>])

Create instance.

	my($banner) = new LinkBanner;

or

	my($logpath) = "/home/foo/public_html/log_path/";
	my($banner) = new LinkBanner($logpath);

or

	my($logpath) = "/home/foo/public_html/log_path/";
	my($query) = new CGI;
	my($banner) = new LinkBanner($logpath,$query);


=item RecordLog(C<$logpath>)

Record referer information to file.

	$banner->RecordLog();

or

	$banner->RecordLog(C<$logpath>);

=item OutBanner(C<$img_file> [,C<$lifetime> ,C<$basetime>])

Output specify banner.
The C<$img_file> specify file of banner image.

The C<$lifetime> specify period of cacheing banner image.
The C<$basetime> specify created time of banner image.
That banner image will be expired at C<$basetime> + C<$lifetime>.
These arguments override default value.

	$banner->OutBanner("/home/foo/public_html/image.gif");

=item OutLinkPage([C<$ra_sorteddata>, C<$rs_template>])

Output the links page that will be made from logged file.
The C<$ra_sorteddata> is reference of logged data array.
The C<$rs_template> is reference of template that is used for page of links.
These arguments override default value.

	$banner->OutLinkPage();

=item deny($filename)

Set filename that is described about denied link address.
It is used for eliminate referer address.
Each line of that file is joined by `|', and eliminate matched address.
Will be escaping  by `\' automatically, if it need.

	$banner->deny("/home/foo/public_html/deny_list.txt");

Example describe denied link address

	http://localhost
	http://www.denied.com


=item debug($debug_opt)

Set debugging option.

	$banner->debug('template');

Message of debugging will be outputted to STDERR.
Also. you can log by redirect STDERR with F<debuglog.pl> or another way.

When you terminate debugging, you can strip debugging code from this module.
All of debugging code are between  line of `#>>>DEBUG_CODE'.
These are able to strip by F<striper.pl>.

=over 4

=item query

Debug about parse of CGI query string.

=item sortedlist

Debug about making of cache link file.

=item ignore

Debug about filtering record.

=item template

Debug about expanding of template.

=item escape

Debug about escaping of the CGI parameters.

=item logfile

Debug about logging.

=back

=item GetTemplateOfLinkPage([C<$rs_buffer>, C<$template>,])

Get template of links page.
In default locate `template.html' as template file in logging file path.

The C<$rs_buffer> specify reference of template buffer that used as argument of  C<OutLinkpage>.
The C<$template> specify filename of template.

=back

=head2 Require modules

F<strict.pm>, F<Carp.pm>, F<File::Basename>, F<URI::Escape>, F<CGI>
F<timestr.pl>, F<regexpr.pl>, F<jcode.pl>

=head2 How to append link banner to web contents

Describe as following to your Web contents.

	<IMG SRC="http://www.any.com/~foo/cgi-bin/linkbanner.cgi?title=title%20of%20your%20page;topic=topic%20of%20your%20page" ALT="Link banner">

You must adjust to your page about `C<title%20of%20your%20page>' and `C<topic%20of%20your%20page>' fields.
Don't use HTML tag. 

=head2 Locate template file of links page

Locate `template.html' as template file in the log file path.

Example template.html ...

	<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01/EN">
	<HTML lang="ja">
	<HEAD>
	<META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=Shift_JIS">
	<TITLE>Links page</TITLE>
	</HEAD>
	<BODY>
	<H1>Links page</H1>
	<HR>
	<TABLE border="1" cellspacing="0">
	<THEAD>
	<TR>
	<TH>Title</TH><TH>topic</TH>
	</TR>
	</THEAD>
	<TBODY>
	<!--
		Following comment statements will be replaced to link list.
		You can customize that.
		These keywords will be handled about specified link banner parameter.
		Like as...

		%referer% : URL of linked content that will reference the link banner.
		%title%   : Title of linked page
		%topic%   : Topic of linked page
	-->
	<!-- __Links_Replace_to_Here__ 
		<TR><TD><A HREF="%referer%">%title%</A></TD><TD>%topic%</TD></TR>
	__Links_Replace_to_Above__ -->
	</TBODY>
	</TABLE>
	<HR>
	</BODY>
	</HTML>

=head1 AUTHOR

Tetsuya Shigetome, E<lt>t_shigetome@muf.biglobe.ne.jpE<gt>.

=head1 HISTORY

=over 4

=item $Revision: 1.6 $

$Log: LinkBanner.pm,v $
Revision 1.6  2002/04/11 15:06:57  T.Shigetome
 o Simplified filter rouitne
 o Change using module from timestr.pl to Timestr.pm

Revision 1.5  2002/03/25 14:43:15  T.Shigetome
Sorry I miss update log in Rev. 1.4.

Fixed bugs

o Sort order is not numericaly.

o Delete new record in FilterSimpleDuplicated.


=item Revision: 1.4 

=item Revision: 1.3

Log: LinkBanner.pm,v
Revision 1.3  2001/07/30 09:38:31  shige

Fixed problem

o Log file grow too big.

=item Revision: 1.2

Log: LinkBanner.pm,v
Revision 1.2  2001/03/13 15:27:06  shige

Fixed bugs

o Duplicate default references.

o Jammed unescaped argument.

o Stricts match expression, so it need change of replace keyword.

Some warning miss match to debug option.


=item Revision: 1.1

Log: LinkBanner.pm,v 
Revision 1.1  2000/12/20 14:41:39  shige

First release of LinkBanenr


=back

=cut

use strict;
$LinkBanner::VERSION = sprintf("%d.%02d", q$Revision: 1.6 $ =~ /(\d+)\.(\d+)/);
use Carp;
use File::Basename;
use File::Copy;
use URI;
use URI::Escape;
use CGI;
use Timestr;
require 'regexpr.pl';
require 'jcode.pl';
require 'filelock.pl';
######################
# Constructer
#
sub new
{
my($pkg,$logpath,$query) = @_;
my($rh_this);
	# comment outed members are reserved for will be used later
	$rh_this =  bless {
#>>>DEBUG_CODE
		'_debug'			=>	{
				'query'			=>	1,
				'sortedlist'	=>	undef,
				'ignore'		=>	undef,
				'template'		=>	undef,
				'escape'		=>	undef,
				'logfile'		=>	undef,
			}, # logfile query ignore template logrecod escape logfile sortedlist
#>>>DEBUG_CODE
		'_now'				=>	$^T,
		'deny'			=>	undef,
		'ignore_agent'	=>	undef,
		'_linkpage'			=>	{
				'_template'			=>	'template.html',
				'_html_buffer'		=>	'',
				'_sorteddata'		=>	[],
				'_life'				=>	12*60*60,		# life time of sorted list
				'_replace_key'		=>	{
						'<!-- __Links_Replace_to_Here__\s+(.+)__Links_Replace_to_Above__ -->'	=>	\&Extend2TemplateLinkinfo,
					},
				'_default_filename'	=> ['index.html',	'index.htm',	'default.htm'	],
			},
		'_banner'			=>	{
				'_life'				=>	30*24*60*60,
				'_mtime'			=>	undef,
				'_base'				=>	$^T,
				'_name'				=>	undef,
				'_data'				=>	undef,
				'_size'				=>	undef,
				'_expire'			=>	undef,
				'_max_age'			=>	undef,
			},
		'_query'			=>	$query,	# reference copy of CGI class instans
		'_log'				=>	{
				'_filename'			=>	\&GetLogFilename,	# it is include path
				'_suffix'			=>	'.log',
				'_path'				=>	$logpath,
				'_field_separator'	=>	'__SePaRaToR__',
				'_record_separator'	=>	'__ReCoRdSePaRaToR__',
				'_field_list'		=>	['entrytime','referer','title','topic'],
				'_field_data'		=>	undef,# [],
#				'_code'				=>	'euc',
				'_code'				=>	'sjis',
				'_conv_code_fields'	=>	['title','topic'],
				'_life'				=>	30*24*60*60,
				'_sortedlist'		=>	'sorted.lst',	# create by this
				'_tempolary'		=>	'links.tmp',
				'_mtime'			=>	undef,	# modified time of sorted list
				'_size'				=>	undef,
				'_limitsize'		=>	32*1024, # 32K byte for limit of monthly log file
				'_filter'			=>	[
											\&FilterIgnoreAgent,
											\&FilterDeny,
											\&FilterEscapedUnsafeChar,
											\&FilterAnchorName,
											\&FilterCGIParameter,
											\&FilterDuplicated,
										],
			},
		'_content_type'		=>	{	# If you want to choose another content type, you MUST be append to below.
				'.jpg'			=>	'image/jpeg',
				'.JPG'			=>	'image/jpeg',
				'.gif'			=>	'image/gif',
				'.GIF'			=>	'image/gif',
			},
	}, $pkg;
	unless( defined($query) )
	{
		$rh_this->{'_query'} = new CGI;
	}
#>>>DEBUG_CODE
	if( defined($rh_this->{'_debug'}{'query'}) )
	{
		push(@{$rh_this->{'_log'}{'_field_list'}},'user_agent','query_string')
	}
#>>>DEBUG_CODE
	return	$rh_this;
}
#######################
# Method autoload
sub AUTOLOAD
{
my($pkg) = shift;
#	return if $LinkBanner::AUTOLOAD =~ m/::DESTROY$/;
my($type) = ref($pkg) || croak "$pkg is not an object";
my($name) = $LinkBanner::AUTOLOAD;

	$name =~ s/.*://;
	unless( exists($pkg->{$name}) )
	{
#>>>DEBUG_CODE
		unless( $name eq 'debug' )
		{
#>>>DEBUG_CODE
			croak "Can't access `$name' field in object of class $type";
#>>>DEBUG_CODE
		}
		else
		{
			foreach( @_ )
			{
				$pkg->{'_debug'}{$_} = 1;
			}
		}
#>>>DEBUG_CODE
	}
	if( @_ )
	{
		return $pkg->{$name} = shift;
	}
	else
	{
		return $pkg->{$name};
	}
}
#	It's a scapegoat for AUTOLOAD.
sub	DESTROY{}
#######################
# Parse query and referer
sub ParseReferer
{
my($pkg) = @_;
my($fieldname);
	if( defined($ENV{'HTTP_REFERER'} && $ENV{'HTTP_REFERER'} !~ m/^\s*$/ ) # It's may be null string
	&& !(defined($ENV{'HTTP_PRAGMA'}) && $ENV{'HTTP_PRAGMA'} =~ m/no-cache/i ))
	{# If clear referer and new request 
	#  then record referer
		$pkg->{'_query'}->param('referer',$ENV{'HTTP_REFERER'});
		if(  defined($pkg->{'_debug'}{'query'}) )
		{
			$pkg->{'_query'}->param('user_agent',$ENV{'HTTP_USER_AGENT'});
			$pkg->{'_query'}->param('query_string',$ENV{'QUERY_STRING'});
		}
		$pkg->{'_query'}->param('entrytime',$pkg->{'_now'});

		# some argument should be convert character set
		foreach $fieldname ( @{$pkg->{'_log'}{'_conv_code_fields'}} )
		{
			$pkg->ConvertString($fieldname);
		}
		foreach( @{$pkg->{'_log'}{'_field_list'}} )
		{
			if( defined($pkg->{'_query'}->param($_)) )
			{
				push(@{$pkg->{'_log'}{'_field_data'}},$pkg->{'_query'}->param($_));
			}
			else
			{
				push(@{$pkg->{'_log'}{'_field_data'}},'');
			}
		}
	}
#>>>DEBUG_CODE
	elsif(  defined($pkg->{'_debug'}{'ignore'}) )
	{
		$pkg->{'_query'}->param('entrytime',$pkg->{'_now'});

		# some argument should be convert code
		foreach $fieldname ( @{$pkg->{'_log'}{'_conv_code_fields'}} )
		{
			$pkg->ConvertString($fieldname);
		}
		foreach( @{$pkg->{'_log'}{'_field_list'}} )
		{
			if( defined($pkg->{'_query'}->param($_)) )
			{
				push(@{$pkg->{'_log'}{'_field_data'}},$pkg->{'_query'}->param($_));
			}
			else
			{
				push(@{$pkg->{'_log'}{'_field_data'}},'');
			}
		}
		my($record_string) = join($pkg->{'_log'}{'_field_separator'},@{$pkg->{'_log'}{'_field_data'}});
		warn $record_string ;
	}
#>>>DEBUG_CODE
}
##################
# Output banner image
sub OutBanner
{
my($pkg,$img_file,$lifetime,$basetime) = @_;
my($name,$path,$suffix);
my($mtime,$expire);

	if( defined($img_file))
	{
		$pkg->LoadImage($img_file);
		if( defined($lifetime) )# Overrided
		{
			$pkg->{'_banner'}{'_life'} = $lifetime;
			if( defined($basetime) )# Overrided
			{
				$pkg->{'_banner'}{'_base'} = $basetime;
			}
		}
	}
	die "Undefined image file" unless defined($pkg->{'_banner'}{'_name'});
	($name,$path,$suffix) = File::Basename::fileparse($pkg->{'_banner'}{'_name'},'\..*');
	{# Get image
		#($main'dev,$main'ino,$main'mode,$main'nl,$main'uid,$main'gid,$main'rdev,$main'size,$main'atime,$main'mtime)

		$mtime = &LocalToGmtDate($pkg->{'_banner'}{'_mtime'});
		$expire = &LocalToGmtDate($pkg->{'_banner'}{'_life'} + $pkg->{'_banner'}{'_base'});
		$pkg->{'_banner'}{'_max_age'} = $pkg->{'_banner'}{'_life'} + $pkg->{'_banner'}{'_base'} - $pkg->{'_now'};
	}
	binmode(STDOUT);
	print	<<"__REPLY__";
Content-type: $pkg->{'_content_type'}{$suffix}; name="$name$suffix"
Content-Transfer-Encoding: binary
Content-Length: $pkg->{'_banner'}{'_size'}
Cache-Control: max-age=$pkg->{'_banner'}{'_max_age'}
Last-Modified: $mtime
Expires: $expire

$pkg->{'_banner'}{'_data'}
__REPLY__
}
##################
# Output link page
sub OutLinkPage
{
my($pkg,$ra_sorteddata,$rs_template) = @_;
my($mtime,$expire,$max_age);

	if( defined($ra_sorteddata) )
	{# IF defined sorted data, then replace 
		$pkg->{'_linkpage'}{'_sorteddata'} = $ra_sorteddata;
	}
	else
	{
		$pkg->UpdateSortedList($pkg->{'_linkpage'}{'_sorteddata'});
	}
	unless( defined($rs_template) )
	{
		$pkg->GetTemplateOfLinkPage();
		$rs_template = \$pkg->{'_linkpage'}{'_html_buffer'};
	}
	foreach( keys(%{$pkg->{'_linkpage'}{'_replace_key'}}) )
	{# Replace keyword in template
		$$rs_template =~ s/$_/&{$pkg->{'_linkpage'}{'_replace_key'}{$_}}($pkg,$1)/sge;
	}
	{
		$mtime = &LocalToGmtDate($pkg->{'_log'}{'_mtime'});
		$expire = &LocalToGmtDate($pkg->{'_linkpage'}{'_life'} + $pkg->{'_log'}{'_mtime'});
		$max_age = $pkg->{'_linkpage'}{'_life'} + $pkg->{'_log'}{'_mtime'} - $pkg->{'_now'};
	}

	print	<<"__LINK_PAGE__";
Content-type: text/html
Cache-Control: max-age=$max_age
Last-Modified: $mtime
Expires: $expire

$$rs_template
__LINK_PAGE__
}
##################
# Make Link table data
sub	Extend2TemplateLinkinfo
{
my($pkg,$format) = @_;
my($field_matchexpr) = join('|',@{$pkg->{'_log'}{'_field_list'}});
my(@split_work,%hash_work,$i,$temp);
my($buffer) = '';
#>>>DEBUG_CODE
warn $format if defined($pkg->{'_debug'}{'template'});
warn $#{$pkg->{'_linkpage'}{'_sorteddata'}} if defined($pkg->{'_debug'}{'template'});
#>>>DEBUG_CODE
	foreach( @{$pkg->{'_linkpage'}{'_sorteddata'}} )
	{
		# clear tempolary buffer
		$i = 0;
		undef %hash_work;

		@split_work = split(/$pkg->{'_log'}{'_field_separator'}/o ,$_);
		foreach( @split_work )
		{
			$hash_work{$pkg->{'_log'}{'_field_list'}[$i]} = $_;
#>>>DEBUG_CODE
warn "$pkg->{'_log'}{'_field_list'}[$i] => $hash_work{$pkg->{'_log'}{'_field_list'}[$i]}" 
	 if defined($pkg->{'_debug'}{'template'});
#>>>DEBUG_CODE
			++$i;
		}
		unless( exists($hash_work{'title'}) && $hash_work{'title'} ne '' )
		{
			$hash_work{'title'} = $hash_work{'referer'};
		}
		$temp = $format;
		$temp =~ s/\%($field_matchexpr)\%/$hash_work{$1}/sge;
#>>>DEBUG_CODE
warn $temp  if defined($pkg->{'_debug'}{'template'});
#>>>DEBUG_CODE
		$buffer .= $temp;
	}
#>>>DEBUG_CODE
warn $buffer  if defined($pkg->{'_debug'}{'template'});
#>>>DEBUG_CODE
	return	$buffer;
}
##################
# Load banner image
sub LoadImage
{
my($pkg,$img_file) = @_;

	if( defined($img_file))
	{
		$pkg->{'_banner'}{'_name'} = $img_file;
	}
	die "Undefined image file" unless defined($pkg->{'_banner'}{'_name'});
	{# Get image
		my(@file_stat) = stat($pkg->{'_banner'}{'_name'});
		$pkg->{'_banner'}{'_mtime'} = $file_stat[9];
		$pkg->{'_banner'}{'_size'} = $file_stat[7];

	# Load data
		open(IMG,$pkg->{'_banner'}{'_name'}) || die "$pkg->{'_banner'}{'_name'} : $!";
		binmode(IMG);
		# clear buffer
		$pkg->{'_banner'}{'_data'} = ();
		while(<IMG>)
		{
			$pkg->{'_banner'}{'_data'} .= $_;
		}
	}
}
#######################
# Record to log file
sub RecordLog
{
my($pkg,$logpath) = @_;
my($record_string);
my($logfile);
	$pkg->ParseReferer();
	if( defined($logpath) )# Overrided
	{
		$pkg->{'_log'}{'_path'} = $logpath;
	}
	if( defined($pkg->{'_log'}{'_field_data'}) && @{$pkg->{'_log'}{'_field_data'}} > 0 )
	{# exist some data and defined log file
		$record_string = join($pkg->{'_log'}{'_field_separator'},@{$pkg->{'_log'}{'_field_data'}});
#>>>DEBUG_CODE
		warn scalar @{$pkg->{'_log'}{'_field_data'}} if defined($pkg->{'_debug'}{'logfile'});
		warn $record_string if defined($pkg->{'_debug'}{'logfile'});
#>>>DEBUG_CODE

		$logfile = $pkg->{'_log'}{'_path'}.&{$pkg->{'_log'}{'_filename'}}($pkg);

		open(LOG,">>$logfile") || die "Can not open $logfile : $!";
		print LOG $record_string.$pkg->{'_log'}{'_record_separator'}."\n";
		close(LOG);
	}
	unless( -s $logfile < $pkg->{'_log'}{'_limitsize'} )# less than limit size
	{
		warn if defined($pkg->{'_debug'}{'logfile'});
		my(@buffer);
		my($tempfile) = sprintf("$pkg->{'_log'}{'_path'}/%08x.%08x",$^T,$$);
		open(LOG,"+<$logfile") || die "Can not open $logfile : $!";
		open(TMP,">$tempfile") || die "Can not create $tempfile : $!";
		&File::Lock('LOG');
		@buffer = <LOG>;
		$pkg->FilterSimpleDuplicated(\@buffer);
		print TMP @buffer;

		close(TMP);
		close(LOG);
		rename($tempfile,$logfile) or die "Fail rename to $logfile from $tempfile: $!";
	}
}
#######################
# Convert statement
sub ConvertString
{
my($pkg,$fieldname) = @_;
my(%special_char) = (
	'<'		=>	'&lt;',
	'>'		=>	'&gt;',
	'&'		=>	'&amp;',
	'\"'	=>	'&quot;',
);
my($match_expr) = '['.join('',keys(%special_char)).']';
	if( defined($pkg->{'_query'}->param($fieldname)) )
	{
		local($LinkBanner'work) = $pkg->{'_query'}->param($fieldname);


		# escape tag
#>>>DEBUG_CODE
warn $match_expr  if defined($pkg->{'_debug'}{'escape'});
#>>>DEBUG_CODE
		# tempolary convert
		&jcode'convert(*LinkBanner'work,'euc');
		$LinkBanner'work =~ s/($match_expr)/$special_char{$1}/geo;

		# convert charset to specify code
		&jcode'convert(*LinkBanner'work,$pkg->{'_log'}{'_code'});

		$pkg->{'_query'}->param($fieldname,$LinkBanner'work);
	}
}
#######################
# Generate monthly file name
sub GetLogFilename
{
my($pkg,$t) = @_;
my(@date);
	if( defined($t) )
	{
#>>>DEBUG_CODE
		warn $t  if defined($pkg->{'_debug'}{'logfile'});
#>>>DEBUG_CODE
		@date = localtime($t);
	}
	else
	{
		@date = localtime($pkg->{'_now'});
	}
# s,m,h,d,m,y,w
	sprintf("%04d%02d$pkg->{'_log'}{'_suffix'}",$date[5]+1900,$date[4]+1);
}
#######################
# Load log
sub LoadLogFile
{
my($pkg,$ra_buffer,$stop,$start) = @_;
my($skip) = 28*24*60*60;
my($current);
my($filename,$beforefilename) = ("","");
my($buffer)  ="";
	unless( defined($stop) )
	{
		$stop = $pkg->{'_now'};
	}
	unless( defined($start) )
	{
		$start = $stop - $pkg->{'_log'}{'_life'};
	}
	for($current = $start ; $skip > 0 ; $current = $current + $skip )
	{
#>>>DEBUG_CODE
warn $pkg->{'_log'}{'_path'} if defined($pkg->{'_debug'}{'logfile'});
#>>>DEBUG_CODE
		$filename = $pkg->{'_log'}{'_path'}.&{$pkg->{'_log'}{'_filename'}}($pkg,$current);
		if( $filename ne $beforefilename )
		{
#>>>DEBUG_CODE
warn $filename if defined($pkg->{'_debug'}{'logfile'});
#>>>DEBUG_CODE
			$beforefilename = $filename;
			if( open(LOG,$filename ) )
			{
				while( <LOG> )
				{
					$buffer .= $_;
				}
				close(LOG);
			}
		}
		if( ($current + $skip) > $stop )
		{
			$skip = $stop - $current;
		}
	}
	# filter specify time range
	@{$ra_buffer} = split(/$pkg->{'_log'}{'_record_separator'}\n*/o,$buffer);
#>>>DEBUG_CODE
warn $#{$ra_buffer}  if defined($pkg->{'_debug'}{'logfile'});
#>>>DEBUG_CODE
	@{$ra_buffer} = reverse sort SortRecordByTimeStamp @{$ra_buffer};
#>>>DEBUG_CODE
warn $#{$ra_buffer}  if defined($pkg->{'_debug'}{'logfile'});
#>>>DEBUG_CODE
	@{$ra_buffer} = grep { if( m/^([0-9]+)$pkg->{'_log'}{'_field_separator'}/o ) {
						($1 > $start && $1 < $stop);} } @{$ra_buffer};
	# Filter referer address blank record
	@{$ra_buffer} = grep { !(m/^([0-9]+)$pkg->{'_log'}{'_field_separator'}\s*$pkg->{'_log'}{'_field_separator'}/o);} @{$ra_buffer};
#>>>DEBUG_CODE
warn $#{$ra_buffer}  if defined($pkg->{'_debug'}{'logfile'});
#>>>DEBUG_CODE
	return	$ra_buffer;
}
#######################
# Sort record by timestamp
sub	SortRecordByTimeStamp
{
my($a_t,$b_t);
	$a =~ m/^(\d+)/;
	$a_t = $1;
	$b =~ m/^(\d+)/;
	$b_t = $1;
	$a_t <=> $b_t;
}
#######################
# Update sorted list
sub UpdateSortedList
{
my($pkg,$ra_buffer) = @_;
my($tempfile) = $pkg->{'_log'}{'_path'}.$pkg->{'_log'}{'_tempolary'};
my($sortedfile) = $pkg->{'_log'}{'_path'}.$pkg->{'_log'}{'_sortedlist'};
my($buffer) = "";
	unless( defined($ra_buffer) )
	{
		$ra_buffer = $pkg->{'_linkpage'}{'_sorteddata'};
	}
	{# check current sorted file
		my(@file_stat) = stat($sortedfile);
		$pkg->{'_log'}{'_mtime'} = 0;
		$pkg->{'_log'}{'_size'} = 0;
		if( @file_stat > 0 )
		{
			$pkg->{'_log'}{'_mtime'} = $file_stat[9];
			$pkg->{'_log'}{'_size'} = $file_stat[7];
		}
	}
#>>>DEBUG_CODE
	if( defined($pkg->{'_debug'}{'sortedlist'}) )
	{
		warn $pkg->{'_now'};
		warn $pkg->{'_log'}{'_life'};
		warn $pkg->{'_log'}{'_mtime'};
		warn $pkg->{'_linkpage'}{'_life'};
	}
#>>>DEBUG_CODE
	if( ($pkg->{'_log'}{'_mtime'} + $pkg->{'_linkpage'}{'_life'}) < $pkg->{'_now'} )# too old
	{
		$buffer = '';
		$pkg->LoadLogFile($ra_buffer);
		# Filter log
		foreach( @{$pkg->{'_log'}{'_filter'}} )
		{
			&{$_}($pkg,$ra_buffer);
#>>>DEBUG_CODE
			warn $#{$ra_buffer}  if defined($pkg->{'_debug'}{'sortedlist'});
#>>>DEBUG_CODE
		}
		{# Create temporaly file
			open(TMP,">$tempfile") || die "Can NOT open $tempfile : $!";
			# file lock
			foreach( @{$ra_buffer} )
			{
				print TMP $_.$pkg->{'_log'}{'_record_separator'}."\n";
			}
			close(TMP);
			# rename file
			rename($tempfile,$sortedfile) or die "Can NOT rename $tempfile to $sortedfile : $!";
		}
		$pkg->{'_log'}{'_mtime'} = $pkg->{'_now'};
	}
	else
	{# Load sorted file
		# defined limit size of  monthly log file
		if( defined($pkg->{'_log'}{'_limitsize'}) )
		{
			$pkg->CutOffLogFile();
		}

		open(TMP,"<$sortedfile") || die "Can NOT open $sortedfile : $!";
		while( <TMP> )
		{
			$buffer .= $_;
		}
		@{$ra_buffer} = split(/$pkg->{'_log'}{'_record_separator'}\n*/o,$buffer);
	}
	return	$ra_buffer;
}
#######################
# Simple Filter same referer
sub FilterSimpleDuplicated
{
my($pkg,$ra_buffer) = @_;
my(%record) = ();
#my(%referer) = ();
#my($referer_adr);
#my($referer_field);
	die "Do NOT specify logbuffer reference" unless(defined($ra_buffer));
#	for( $referer_field = 0 ; $pkg->{'_log'}{'_field_list'}[$referer_field] ne 'referer' ; ++$referer_field ){}

#	foreach( @{$ra_buffer} )
#	{
#		$referer_adr = (split(/$pkg->{'_log'}{'_field_separator'}/o,$_))[$referer_field];
#		# last one is newest. So It's overwrited.
#		$referer{$referer_adr} = $_;
#	}
	$pkg->FilterIgnoreAgent($ra_buffer, );
	# Filter same record 
	foreach( @{$ra_buffer} )
	{
		m/^\d+(.+)$/;
		# last one is newest. So It's overwrited.
		$record{$1} = $_;
	}
	@{$ra_buffer} = sort SortRecordByTimeStamp values(%record);

	%record = ();
my($i, @field, @record);
	$i = 0;
	foreach( @{$pkg->{'_log'}{'_field_list'}} )
	{
		if( $_ eq 'referer' )
		{
			push(@field, $i);
		}
		elsif( $_ eq 'title' )
		{
			push(@field, $i);
		}
		elsif( $_ eq 'topic' )
		{
			push(@field, $i);
		}
		++$i;
	}
	foreach( @{$ra_buffer} )
	{
		@record = split(/$pkg->{'_log'}{'_field_separator'}/o,$_);
		# last one is newest. So It's overwrited.
		$record{$record[$field[0]].$record[$field[1]].$record[$field[2]]} = $_;
	}
	@{$ra_buffer} = sort SortRecordByTimeStamp values(%record);

	# We MUST keep sort order by time stamp in logfile.
#	@{$ra_buffer} = sort SortRecordByTimeStamp values(%referer);
	return	$ra_buffer;
}
#######################
# Filter same link
sub FilterDuplicated
{
my($pkg,$ra_buffer) = @_;
my(%referer) = ();
my($referer_field);
my($referer_adr);
my(@temp) = ();
my($default_filename_matchexpr);

	undef	@temp;
	push(@temp,&Regexpr::Escape(@{$pkg->{'_linkpage'}{'_default_filename'}}));
	$default_filename_matchexpr = join('|',@temp);
#>>>DEBUG_CODE
warn $default_filename_matchexpr  if defined($pkg->{'_debug'}{'sortedlist'});
#>>>DEBUG_CODE

	unless(defined($ra_buffer))
	{
		$ra_buffer = $pkg->{'_linkpage'}{'_sorteddata'};
	}
	for( $referer_field = 0 ; $pkg->{'_log'}{'_field_list'}[$referer_field] ne 'referer' ; ++$referer_field ){}

	undef	@temp;
	foreach( @{$ra_buffer} )
	{
		$referer_adr = (split(/$pkg->{'_log'}{'_field_separator'}/o,$_))[$referer_field];
		$referer_adr = &uri_escape($referer_adr,"\x00-\x1f\x7e-\xff");
		$referer_adr =~ tr/A-Z/a-z/;
#>>>DEBUG_CODE
		my($u1) = URI->new($referer_adr);
warn $referer_adr  if defined($pkg->{'_debug'}{'sortedlist'});
#>>>DEBUG_CODE
		unless( exists($referer{$referer_adr}) || exists($referer{"$referer_adr/"}) )
		{# New record
			my($it_first) = 1;
			if( $referer_adr =~ m|/($default_filename_matchexpr)$|o )
			{# Default file
				if( $1 eq '' )
				{
					foreach(@{$pkg->{'_linkpage'}{'_default_filename'}})
					{
						if( exists($referer{"$referer_adr$_"}) )
						{
							$it_first = 0;
						}
#>>>DEBUG_CODE
						else
						{
							warn $referer_adr.$_  if defined($pkg->{'_debug'}{'sortedlist'});
						}
#>>>DEBUG_CODE
					}
				}
				else
				{
					if( exists($referer{"$`/"}) || exists($referer{"$`"}) )
					{
						$it_first = 0;
					}
#>>>DEBUG_CODE
					else
					{
						warn $referer_adr  if defined($pkg->{'_debug'}{'sortedlist'});
					}
#>>>DEBUG_CODE
				}
			}
			else
			{
				if( $referer_adr !~ m|/[^/]+\.[^/]+$| || $u1->path eq '' )
				{# If it may be directory name
					my($separator) = '';
					# if it not finished directory separator then append
					$separator = '/' if $referer_adr =~ m|[^/]$|;

#>>>DEBUG_CODE
warn $referer_adr  if defined($pkg->{'_debug'}{'sortedlist'});
#>>>DEBUG_CODE

					if( exists($referer{"$referer_adr$separator"}) )
					{
						$it_first = 0;
					}
					else
					{
						foreach( @{$pkg->{'_linkpage'}{'_default_filename'}} )
						{
#>>>DEBUG_CODE
warn "$referer_adr$separator$_"  if defined($pkg->{'_debug'}{'sortedlist'});
#>>>DEBUG_CODE
							if( exists($referer{"$referer_adr$separator$_"}) )
							{
								$it_first = 0;
								last;
							}
						}
					}
				}
#>>>DEBUG_CODE
				else
				{
warn "$referer_adr"  if defined($pkg->{'_debug'}{'sortedlist'});
				}
#>>>DEBUG_CODE
			}
			if( $it_first )
			{
#>>>DEBUG_CODE
warn $referer_adr  if defined($pkg->{'_debug'}{'sortedlist'});
#>>>DEBUG_CODE
				$referer{$referer_adr} = 1;
				push(@temp,$_);
			}
		}
	}
	@{$ra_buffer} = @temp;
	return	$ra_buffer;
}
#######################
# Filter denied entry
sub	FilterDeny
{
my($pkg,$ra_buffer,$deny) = @_;
my(@deny_array);
my($match_expr);

	$pkg->{'deny'} = $deny if( defined($deny) );
	if( defined($pkg->{'deny'}) )
	{# load deny file
		$deny = $pkg->{'deny'};
		open(LOG,"<$deny") || die "$deny : $!";
		@deny_array = <LOG>;
		close(LOG);
		chomp @deny_array;
		&Regexpr::Escape(\@deny_array);
		$match_expr = join('|',@deny_array);
#>>>DEBUG_CODE
		warn $match_expr  if defined($pkg->{'_debug'}{'sortedlist'});
#>>>DEBUG_CODE
		@{$ra_buffer} = grep {!(m/($match_expr)/so);} @{$ra_buffer};
	}
	return	$ra_buffer;
}
#######################
# Filter ignore user-agent entry
#
#Note.
# This is almost same to FilterDeny.
# But, This is sepalated. I think this is not suited.
# It may be restructured or customized by user.
sub	FilterIgnoreAgent
{
my($pkg,$ra_buffer,$ignore_agent) = @_;
my(@agent_array);
my($match_expr);

	$pkg->{'ignore_agent'} = $ignore_agent if( defined($ignore_agent) );
	if( defined($pkg->{'ignore_agent'}) )
	{# load deny file
		$ignore_agent = $pkg->{'ignore_agent'};
		open(LOG,"<$ignore_agent") || die "$ignore_agent : $!";
		@agent_array = <LOG>;
		close(LOG);
		chomp	@agent_array;
		&Regexpr::Escape(\@agent_array);
		$match_expr = join('|',@agent_array);
#>>>DEBUG_CODE
		warn $match_expr  if defined($pkg->{'_debug'}{'sortedlist'});
#>>>DEBUG_CODE
		@{$ra_buffer} = grep {!(m/($match_expr)/so);} @{$ra_buffer};
	}
	return	$ra_buffer;
}
#######################
# Filter escaped to ? unsafe character code by user-agent
sub	FilterEscapedUnsafeChar
{
my($pkg,$ra_buffer) = @_;
	@{$ra_buffer} = grep {!(m/(\.\.|\?\?)/o);} @{$ra_buffer};
	return	$ra_buffer;
}
##################
# Filter parameter
sub	FilterCGIParameter
{
my($pkg,$ra_buffer) = @_;
my(%referer) = ();
my($referer_field);
my(@temp) = ();
my(@record);
my($match_expr);
	unless(defined($ra_buffer))
	{
		$ra_buffer = $pkg->{'_linkpage'}{'_sorteddata'};
	}
	foreach( @{$pkg->{'_log'}{'_field_list'}} )
	{
		if( $_ eq 'referer' )
		{
			push(@temp,'[^?]+)\?.*(');
		}
		else
		{
			push(@temp,'.*');
		}
	}
	$match_expr = '('.join($pkg->{'_log'}{'_field_separator'},@temp).')';
#>>>DEBUG_CODE
	warn $#{$pkg->{'_log'}{'_field_list'}}  if defined($pkg->{'_debug'}{'sortedlist'});
	warn $match_expr  if defined($pkg->{'_debug'}{'sortedlist'});
#>>>DEBUG_CODE
	foreach( @{$ra_buffer} )
	{
#>>>DEBUG_CODE
		unless( defined($pkg->{'_debug'}{'sortedlist'}) )
		{
#>>>DEBUG_CODE
			s/^$match_expr/$1$2/o;
#>>>DEBUG_CODE
		}
		else
		{
			warn;
			warn $_;
			if( s/^$match_expr/$1$2/o )
			{
				warn $_;
			}
		}
#>>>DEBUG_CODE
	}
	return	$ra_buffer;
}
##################
# Filter anchor name
sub	FilterAnchorName
{
my($pkg,$ra_buffer) = @_;
my(%referer) = ();
my($referer_field);
my(@temp) = ();
my(@record);
my($match_expr);
	unless(defined($ra_buffer))
	{
		$ra_buffer = $pkg->{'_linkpage'}{'_sorteddata'};
	}
	foreach( @{$pkg->{'_log'}{'_field_list'}} )
	{
		if( $_ eq 'referer' )
		{
			push(@temp,'[^#]+)\#.*(');
		}
		else
		{
			push(@temp,'.*');
		}
	}
	$match_expr = '('.join($pkg->{'_log'}{'_field_separator'},@temp).')';
#>>>DEBUG_CODE
	warn $#{$pkg->{'_log'}{'_field_list'}}  if defined($pkg->{'_debug'}{'sortedlist'});
	warn $match_expr  if defined($pkg->{'_debug'}{'sortedlist'});
#>>>DEBUG_CODE
	foreach( @{$ra_buffer} )
	{
#>>>DEBUG_CODE
		unless( defined($pkg->{'_debug'}{'sortedlist'}) )
		{
#>>>DEBUG_CODE
			s/^$match_expr/$1$2/o;
#>>>DEBUG_CODE
		}
		else
		{
			warn $_;
			if( s/^$match_expr/$1$2/o )
			{
				warn $_;
			}
		}
#>>>DEBUG_CODE
	}
	return	$ra_buffer;
}
##################
# Get Link page template
sub GetTemplateOfLinkPage
{
my($pkg,$rs_buffer,$template) = @_;
my($filename);
	if( defined($rs_buffer) )
	{
		if( defined($template) )
		{
			$pkg->{'_linkpage'}{'_template'} = $template;
		}
	}
	else
	{
		$rs_buffer = \$pkg->{'_linkpage'}{'_html_buffer'};
	}
	$filename = $pkg->{'_log'}{'_path'}.$pkg->{'_linkpage'}{'_template'};
	open(TMP,"<$filename")  || die "Can NOT open html template $filename :$!";
	while(<TMP>)
	{
		$$rs_buffer .= $_;
	}
}
##################
# Cut off log file
sub CutOffLogFile
{
my($pkg) = @_;
my($filename) = $pkg->{'_log'}{'_path'}.&{$pkg->{'_log'}{'_filename'}}($pkg);
my($sortedfile) = $pkg->{'_log'}{'_path'}.$pkg->{'_log'}{'_sortedlist'};
my($tempfile) = sprintf("%07x.cut",$$);
my(@buffer) = ();
my(@sorted_buffer) = ();
my($start_time);
	if( -s $filename > $pkg->{'_log'}{'_limitsize'} )
	{
		$start_time = (stat($sortedfile))[9];
		$pkg->LoadLogFile(\@buffer,$pkg->{'_now'},$start_time);
		copy($sortedfile,$tempfile) || die "Can NOT copy file $sortedfile to $tempfile : $!";

		# Append new record
		open(TMP,">>$tempfile") || die "Can NOT open $tempfile: $!";
		foreach( @buffer )
		{
			print TMP $_.$pkg->{'_log'}{'_record_separator'}."\n";
		}
		close(TMP);
		rename($tempfile,$filename) || die "Can NOT rename $tempfile to $filename : $!";
	}
}
1;
